from adsb_tools.utils import requests_utils

class Receiver:
    def __init__(self, base_adsb_url):
        self.url = f'{base_adsb_url}/data/receiver.json'
        receiver = requests_utils.get_api(self.url)
        self.version = receiver['version']
        self.refresh = receiver['refresh']
        self.history = receiver['history']
        self.version = receiver['version']
        self.lat = receiver['lat']
        self.lon = receiver['lon']
        self.latitude = receiver['lat']
        self.longitude = receiver['lon']



import math
import numbers
from adsb_tools.utils import requests_utils, spatial_utils
from pprint import pprint

EARTH_RADIUS_KM = 6371.0

RADIUS_EARTH_FEET = 20925524.9  # Radius of the Earth in feet
RADIUS_EARTH_MILES = 3958.8  # Radius of the Earth in miles
RADIUS_EARTH_KM = 6371.0  # Radius of the Earth in kilometers
RADIUS_EARTH_NM = 3440.1  # Radius of the Earth in nautical miles
RADIUS_EARTH_METERS = 6371000.0  # Radius of the Earth in meters

class Aircraft:
    def __init__(self, base_adsb_url, base_latitude=None, base_longitude=None, flightaware_api_key=None):
        self.base_adsb_url = base_adsb_url
        self.earth_radius_km = EARTH_RADIUS_KM
        aircraft_list = self.get_aircraft_list()
        self.aircraft_list = aircraft_list
        self.nearest_aircraft = {}
        self.base_latitude = None
        self.base_longitude = None
        self.flightaware_api_key = flightaware_api_key
        if (base_latitude and base_longitude):
            self.base_latitude = base_latitude
            self.base_longitude = base_longitude
            self.augment_aircraft_list()
            self.set_nearest_aircraft()


    def augment_aircraft_list(self):
        """
        Adds additional options to the aircraft list and sets it as the new `aircraft_list`.
        This method does not exist as part of init, so that it can be called separately by whatever has instantiated the
        Aircraft class
        """
        aircraft_list = self.add_aircraft_options()
        self.aircraft_list = aircraft_list


    def set_nearest_aircraft(self):
        """
        Sets the nearest aircraft to the current location and adds links to various databases.

        The function creates a copy of the first aircraft in the aircraft list, adds links to the HEXDB and ADSB databases,
        and sets the result as the nearest aircraft.

        Returns:
            None
        """
        # dict creates a copy of the dictionary. aircraft_list remains unaffected.
        nearest_aircraft = dict(self.aircraft_list[0])
        icao_24 = nearest_aircraft['hex']
        nearest_aircraft['hexdb'] = {
            'aircraft_url': f'https://hexdb.io/api/v1/aircraft/{icao_24}',
            'conversion_url': f'https://hexdb.io/hex-reg?hex={icao_24}'
        }
        nearest_aircraft['adsb_db'] = {
            'aircraft_url': f'https://api.adsbdb.com/v0/aircraft/{icao_24}',
            'conversion_url': f'https://api.adsbdb.com/v0/mode-s/{icao_24}'
        }

        self.nearest_aircraft = nearest_aircraft


    def retrieve_external_aircraft_options(self):
        """
        Retrieves external flight information and aircraft image for the nearest aircraft.
        Then updates the `nearest_aircraft` dictionary with the aircraft image URL and the merged flight information.
        """
        nearest_aircraft = self.nearest_aircraft
        icao_24 = nearest_aircraft['icao_24']
        aircraft_image = Aircraft.get_aircraft_image(icao_24)
        hex_db_options = Aircraft.get_hex_db_flight(icao_24)
        adsb_db_options = Aircraft.get_adsb_db_flight(icao_24)

        flight_options = {}
        if adsb_db_options:
            flight_options = adsb_db_options
        elif hex_db_options:
            flight_options = hex_db_options

        nearest_aircraft.update(flight_options)
        nearest_aircraft['image'] = aircraft_image
        nearest_aircraft['identity'] = {
            **flight_options,
            'image': aircraft_image
        }

        if ('flightaware_urls' in nearest_aircraft and bool(nearest_aircraft['flightaware_urls'])):
            live_url = f"https://flightaware.com/live/flight/{flight_options['registration']}"
            nearest_aircraft['flightaware_urls']['live'] = live_url

        self.nearest_aircraft = nearest_aircraft

    
    def set_flightaware_ident(self):
        headers = {'x-apikey': self.flightaware_api_key}
        nearest_aircraft = self.nearest_aircraft
        registration = nearest_aircraft['identity']['registration']

        aeroapi_url = f'https://aeroapi.flightaware.com/aeroapi/flights/{registration}'
        json_obj = requests_utils.get_api(url=aeroapi_url, headers=headers)

        current_flight = {}
        if ('flights' in json_obj and len(json_obj['flights']) != 0):
            filtered_data = [d for d in json_obj['flights'] if not d['status'].lower().startswith(('scheduled', 'arrived'))]
            current_flight = filtered_data[0]
        
        if bool(current_flight):
            nearest_aircraft['flightaware'] = current_flight
        
        return current_flight


    def map_static_aircraft_options(self, stored_aircraft):
        """
        Maps static aircraft information from a stored aircraft dictionary to the nearest aircraft dictionary.
        """
        keys_to_map = [
            'identity',
            'flightaware',
            'flightaware_urls'
        ]
        for key in keys_to_map:
            if key in stored_aircraft:
                self.nearest_aircraft[key] = stored_aircraft[key]


    def get_aircraft_list(self):
        """
        Get the aircraft messages and returns as list
        """
        aircraft_url = f'{self.base_adsb_url}/data/aircraft.json'
        json_obj = requests_utils.get_api(aircraft_url)
        result = json_obj['aircraft']

        return result


    def add_aircraft_options(self):
        """
        Adds additional data to aircraft based on aircraft properties
        """
        aircraft_list_with_options = []
        for aircraft in self.aircraft_list:
            if 'flight' in aircraft:
                aircraft['flight'] = aircraft['flight'].strip()
            if Aircraft.get_has_coordinates(aircraft):
                new_aircraft = aircraft.copy()
                aircraft_lat = new_aircraft['lat']
                aircraft_lon = new_aircraft['lon']

                distances = spatial_utils.calculate_distances(
                    self.base_latitude, self.base_longitude, aircraft_lat, aircraft_lon
                )
                degrees = spatial_utils.get_degrees(
                    self.base_latitude, self.base_longitude, aircraft_lat, aircraft_lon
                )
                direction = spatial_utils.get_cardinal_direction(degrees)

                new_aircraft['distance'] = {
                    **distances,
                    'degrees': degrees,
                    'direction': direction,
                }
                new_aircraft['icao_24'] = aircraft['hex']
                movement = {}
                if ('track' in aircraft and isinstance(aircraft['track'], numbers.Number)):
                    movement = {
                        'track': aircraft['track'],
                        'direction': spatial_utils.get_cardinal_direction(aircraft['track'])
                    }
                new_aircraft['movement'] = movement

                mode_s = aircraft['hex']
                redirect_url = f"https://flightaware.com/live/modes/{mode_s}/redirect"
                if ('flight' in aircraft and bool(aircraft['flight'])):
                    redirect_url = f"https://flightaware.com/live/modes/{mode_s}/ident/{aircraft['flight']}/redirect"

                new_aircraft['flightaware_urls'] = {
                    'redirect': redirect_url
                }

                aircraft_list_with_options.append(new_aircraft)
            else:
                aircraft_list_with_options.append(aircraft)

        return Aircraft.sort_by_distance(aircraft_list_with_options)


    @staticmethod
    def get_hex_db_flight(icao_24):
        """
        Get values from HexDB
        """
        hex_db_url = f'https://hexdb.io/api/v1/aircraft/{icao_24}'
        hex_db_obj = requests_utils.get_api(hex_db_url)

        mapped_keys = {
            'icao_type_code': 'ICAOTypeCode',
            'country_iso': None,
            'country_name': None,
            'manufacturer': 'Manufacturer',
            'mode_s': 'ModeS',
            'operator_flag_code': 'OperatorFlagCode',
            'owner': 'RegisteredOwners',
            'registration': 'Registration',
            'type': 'Type'
        }

        return requests_utils.map_keys(hex_db_obj, mapped_keys)


    @staticmethod
    def get_adsb_db_flight(icao_24):
        """
        Get values from ADSB DB
        """
        adsb_db_url = f'https://api.adsbdb.com/v0/aircraft/{icao_24}'
        adsb_db_obj = requests_utils.get_api(adsb_db_url)
        response = adsb_db_obj['response']

        result = {}
        if (type(response) is dict and 'aircraft' in response):
            adsb_aircraft = response['aircraft']
            mapped_keys = {
                'icao_type_code': 'icao_type',
                'manufacturer': 'manufacturer',
                'country_iso': 'registered_owner_country_iso_name',
                'country_name': 'registered_owner_country_name',
                'mode_s': 'mode_s',
                'operator_flag_code': 'registered_owner_operator_flag_code',
                'owner': 'registered_owner',
                'registration': 'registration',
                'type': 'type'
            }
            result = requests_utils.map_keys(adsb_aircraft, mapped_keys)
        return result
    

    @staticmethod
    def get_has_coordinates(aircraft = {}) -> bool:
        """
        Determine whether an aircraft has coordinates
        """
        return True if aircraft.get("lat") is not None and aircraft.get("lon") is not None else False


    @staticmethod
    def get_aircraft_image(aircraft_hex):
        """
        Gets the image of an aircraft, given its hex. If no image found, return
        empty dictionary.
        """
        headers = {
            'User-Agent': 'My Unique User Agent'
        }
        planespotter_url = f'https://api.planespotters.net/pub/photos/hex/{aircraft_hex}'
        json_obj = requests_utils.get_api(planespotter_url, headers = headers)

        image = {}
        if (json_obj['photos'] and json_obj['photos'][0] and bool(json_obj['photos'][0])):
            photo_attributes = json_obj['photos'][0]
            thumbnail = photo_attributes['thumbnail']
            thumbnail_large = photo_attributes['thumbnail_large']
            target_photo = thumbnail if thumbnail_large is None else thumbnail_large
            image = {
                'height': target_photo['size']['height'],
                'width': target_photo['size']['width'],
                'src': target_photo['src'],
                'url': planespotter_url
            }

        return image
    

    @staticmethod
    def sort_by_distance(data):
        """
        Sort a list of dictionaries based on their "distance" values, with dictionaries that have a defined
        "distance" value appearing first.
        """
        # Split dictionaries into those with a distance key and those without
        with_distance = []
        without_distance = []
        for d in data:
            if "distance" in d:
                with_distance.append(d)
            else:
                without_distance.append(d)

        # Sort dictionaries with distance by their distance values
        with_distance_sorted = sorted(with_distance, key=lambda x: x.get("distance").get("km", float("inf")))

        # Combine the two lists, with dictionaries with distance at the beginning
        return with_distance_sorted + without_distance



"""
This module provides several utility functions for making HTTP requests and processing JSON responses.

Functions:
- `call_url(url: str, timeout: int = 5, headers: dict = {}) -> requests.Response`: Sends an HTTP GET request to the
  specified URL with the given timeout.
- `get_api(url, timeout = 5, headers = {})`: Makes a GET request to a Rest API and returns a dictionary or list.
- `map_keys(original_dict, mapped_keys)`: Takes the values from one dictionary and returns a new dictionary with the
  same values, but with different key names.

Example Usage:
--------------
>>> response = call_url('https://api.github.com/users/octocat/repos')
>>> content = get_api('https://api.github.com/users/octocat/repos')
>>> new_dict = map_keys({'a': 1, 'b': 2}, {'c': 'a', 'd': None})
"""

import json
from typing import Dict
import time
import requests

def call_url(url: str, timeout: int = 5, headers: Dict[str, str] = {}) -> requests.Response:
    """
    Sends an HTTP GET request to the specified URL with the given timeout.

    Parameters:
        url (str): The URL to send the request to.
        timeout (int): The maximum number of seconds to wait for a response.

    Returns:
        requests.Response: A `Response` object containing information about the response from the server,
            or `None` if an error occurs.
    """
    start_time = time.time() * 1000
    try:
        print(f"{url}: attempting request...")
        response = requests.get(url=url, timeout=timeout, headers=headers)
        response.raise_for_status()
    except requests.exceptions.Timeout:
        print(f"{url}: timeout error occurred.")
    except requests.exceptions.HTTPError as err:
        print(f"{url}: HTTP error occurred: {err}")
    except requests.exceptions.RequestException as err:
        print(f"{url}: An error occurred: {err}")
    finally:
        end_time = time.time() * 1000
        print(f'{url}: request completed, {end_time - start_time:.0f}ms')
    return response


def get_api(url: str, timeout: int = 5, headers = {}):
    """
    makes a GET request to a Rest API and returns a dictionary or list
    """
    result = call_url(url, timeout, headers)
    content = json.loads(result.content)
    return content


def map_keys(original_dict: Dict, mapped_keys: Dict) -> Dict:
    """
    Takes the values from one dictionary and returns a new dictionary with the
    same values, but with different key names
    """
    new_dict = {}

    for key in mapped_keys:
        mapped_key = mapped_keys[key]
        if mapped_key is None:
            new_dict[key] = None
        elif mapped_key in original_dict:
            new_dict[key] = original_dict[mapped_key]

    return new_dict



import datetime
import pytz

def format_datetime(datetime_str, timezone):
    dt = datetime.datetime.strptime(datetime_str, '%Y-%m-%dT%H:%M:%SZ')
    tz = pytz.timezone(timezone)
    localized_dt = tz.localize(dt, is_dst=None)
    formatted_dt = localized_dt.strftime('%d %b, %I:%M %p %Z')
    return formatted_dt



import csv
import os

class WakeVortexCategories:
    """
    A class that loads wake vortex categories from a CSV file and provides access to them through various methods.

    The CSV file should have the following columns: Category, TC, CA, ICAO, Description.

    Attributes:
        categories (list): A list of dictionaries representing each category.
        category_dict (dict): A dictionary mapping each category name to its corresponding dictionary.
    """
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._categories = []
            cls._instance._categories_dict = {}
            cls._instance._category_dict = {}
            cls._instance._parse_csv()
        return cls._instance

    @property
    def categories(self):
        """
        Returns the list of categories as a list of dictionaries.

        Returns:
            list: A list of dictionaries representing each category.
        """
        return self._categories

    @property
    def category_dict(self):
        """
        Returns a dictionary mapping category names to their corresponding dictionaries.

        Returns:
            dict: A dictionary mapping category names to their corresponding dictionaries.
        """
        return self._category_dict

    def get_category_dict(self, category):
        """
        Returns the dictionary corresponding to the given category name.

        Args:
            category (str): The name of the category.

        Returns:
            dict: A dictionary representing the category.
        """
        return self._category_dict.get(category.lower())

    def _parse_csv(self):
        """
        Parses the CSV file and populates the internal category dictionaries.
        """
        # get the directory of the current script (foo.py)
        current_dir = os.path.dirname(os.path.abspath(__file__))

        # construct the path to the csv file
        csv_file_path = os.path.join(current_dir, 'static', 'wake_vortex_categories.csv')
        with open(csv_file_path, newline='') as csvfile:
            reader = csv.DictReader(csvfile)
            for row in reader:
                category = row['Category']
                tc = row['TC']
                ca = row['CA']
                icao = row['ICAO']
                description = row['Description']
                category_dict = {
                    'tc': int(tc) if tc != "" else None,
                    'ca': int(ca) if ca != "" else None,
                    'icao': icao,
                    'category': category,
                    'description': description
                }
                self._categories.append(row)
                self._category_dict[category.lower()] = category_dict



"""
This module provides a function to retrieve the name of the timezone for a given set of coordinates.

The function `get_timezone_name` uses the TimezoneFinder library to determine the timezone
at a specific latitude and longitude coordinate pair. It returns the name of the timezone as a string.

Example Usage:
--------------
>>> get_timezone_name(37.7749, -122.4194)
'America/Los_Angeles'
"""

from timezonefinder import TimezoneFinder

def get_timezone_name(latitude, longitude):
    """
    Gets the name of the timezone based on coordinates
    """
    tf = TimezoneFinder()
    timezone_name = tf.timezone_at(lat=latitude, lng=longitude)

    return timezone_name



"""
Spatial Utilities
"""

import math
import numbers

RADIUS_EARTH_FEET = 20925524.9  # Radius of the Earth in feet
RADIUS_EARTH_MILES = 3958.8  # Radius of the Earth in miles
RADIUS_EARTH_KM = 6371.0  # Radius of the Earth in kilometers
RADIUS_EARTH_NM = 3440.1  # Radius of the Earth in nautical miles
RADIUS_EARTH_METERS = 6371000.0  # Radius of the Earth in meters


def calculate_distances(lat1, lon1, lat2, lon2):
    """
    Calculates the great-circle distance between two points on the Earth's surface
    using the Haversine formula.

    Args:
        lat1 (float): Latitude of the first point in degrees
        lon1 (float): Longitude of the first point in degrees
        lat2 (float): Latitude of the second point in degrees
        lon2 (float): Longitude of the second point in degrees

    Returns:
        float: Distance between the two points in kilometers
    """

    # Convert coordinates to radians
    lat1, lon1, lat2, lon2 = map(math.radians, [lat1, lon1, lat2, lon2])

    # Haversine formula
    dlat = lat2 - lat1
    dlon = lon2 - lon1
    square_half_chord = math.sin(dlat/2)**2 + math.cos(lat1) * math.cos(lat2) * math.sin(dlon/2)**2
    angular_distance = 2 * math.atan2(math.sqrt(square_half_chord), math.sqrt(1-square_half_chord))

    # return distance
    return {
        "ft": RADIUS_EARTH_FEET * angular_distance,
        "mi": RADIUS_EARTH_MILES * angular_distance,
        "km": RADIUS_EARTH_KM * angular_distance,
        "nm": RADIUS_EARTH_NM * angular_distance,
        "m": RADIUS_EARTH_METERS * angular_distance
    }



def get_degrees(base_lat, base_lon, dest_lat, dest_lon):
    """Calculate the degrees from one coordinate to another."""
    # Calculate the difference between the latitudes and longitudes
    lat_diff = dest_lat - base_lat
    lon_diff = dest_lon - base_lon

    # Calculate the angle between the two points in radians
    angle = math.atan2(lon_diff, lat_diff)

    # Convert the angle from radians to degrees
    return math.degrees(angle)


def get_cardinal_direction(degrees):
    """Calculate the direction from one coordinate to another."""
    # Convert the angle to a compass direction
    directions = ['N', 'NE', 'E', 'SE', 'S', 'SW', 'W', 'NW']
    idx = round(degrees / (360.0 / len(directions))) % len(directions)
    return directions[idx]
